module edetect_io_ml
  
  use errorHandling
  use param
  use dal

  implicit none

  private

  public cp_head, get_files, co_image, co_bgima, co_expima, co_mask, rawcoords, impix2equ
  
! MAX_IMA              max number of images
! MAX_INS              max number of different instruments 
! MAX_BANDS            max number of energy bands

  integer, parameter, public :: max_ima = 60
  integer, parameter, public :: max_ins = 12
  integer, parameter, public :: max_bands = 6

      integer, parameter,public :: max_fit = 10        
      integer, parameter,public :: max_par = max_fit*max_ima+max_fit*3

! NUM_INSTR            number of instruments
! NUM_IMAGE(MAX_INS)   number of images per instrumet
! NIM                  total number of images

  integer, public :: num_instr, nim
  integer, dimension(max_ins), public :: num_image       
  integer, dimension(max_ima), public :: pimin, pimax

! IMAGE                input images
! BG_IMAGE             background images
! EXP_IMAGE            exposure images

  integer, dimension(:,:,:), allocatable, public  :: image
  real, dimension (:,:,:), allocatable, public :: exp_image, bg_image
  logical, dimension(:,:,:), allocatable, public :: mask_image
  logical, dimension(max_ima), public :: OotCorrDone  
  

  integer, dimension(2), public :: npix

! IP(MAX_INS,MAX_BANDS)pointer field to manage images

  integer, dimension(max_ins, max_bands), public :: ip
  integer, dimension(max_ins), public :: expip

  real(kind=single), dimension(max_ima), public :: ecf

  real(kind=single), public :: xcen1, ycen1
  real(kind=double), dimension(max_ins), public :: ra_pnt, dec_pnt, pa_pnt
  real(kind=double), public :: crpix1,crpix2,crval1,crval2, &
                               crdelt1,crdelt2,crota2
  real(kind=single), dimension(3), public :: xoff, yoff
  character(len=30), public :: submode
  character(len=20), public :: datetime
  character(8) :: m1filt,m2filt,pnfilt
  character(30) :: m1sub,m2sub,pnsub
  real(kind=single) :: pnexpos,m1expos,m2expos

  real(kind=single), dimension(12,max_ima), public :: ontime


  real(kind=double) :: pi=3.14159265358979323846d0

contains

  !---------------------------------------------------------------------
  subroutine cp_attr (set1, set2, arr, attr)
  !---------------------------------------------------------------------
    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT), intent(IN) :: arr
    character(len=*), dimension(:), intent(IN) :: attr

    integer i

    do i=1,size(attr)
      if (hasAttribute(set1,attr(i))) then
        call setAttribute(set2,attribute(set1,attr(i)))
      else 
        if (hasAttribute(arr,attr(i))) then
          call setAttribute(set2,attribute(arr,attr(i)))
        end if
      end if
    end do

  end subroutine cp_attr

  !---------------------------------------------------------------------
  subroutine cp_head (set1, set2, content, xid_arr, mod)
  !---------------------------------------------------------------------

    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT) :: arr
    character(len=*), intent (IN) :: content
    integer, dimension(:), intent(in) :: xid_arr
    integer, intent(IN) :: mod
    character(len=2), dimension(3) :: strInst
    character(len=8)  :: strEmin,strEmax 
    character(len=1)  :: strBand 
    character(len=80) :: strComment
    integer i,j,j0,jj,ipoint
    integer :: num_instr0, n_instr
    integer, dimension(max_ins) :: num_image0
    logical :: xidflg    

    data strInst /"PN","M1","M2"/

    arr = array (set1,"PRIMARY",read)

    if (mod >= 1 .and. mod <=4 ) then
      call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
    end if

    if (mod == 2) then
      call cp_attr(set1,set2,arr,(/"BAND"/) )
    end if

    !if (mod == 3) then
    !  call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
    !end if
    
    if (mod < 3) then  
      call cp_attr(set1,set2,arr, &
      (/ "TELESCOP","DATAMODE","FILTER  ","OBS_MODE","OBS_ID  ","EXP_ID  ", &
         "EXPIDSTR","ORIGIN  ","OBJECT  ","RA_PNT  ","DEC_PNT ","PA_PNT  ", &
         "CRPIX1  ","CRPIX2  ","CRVAL1  ","CRVAL2  ","CDELT1  ", &
         "CDELT2  ","CTYPE1  ","CTYPE2  ","SUBMODE ","RADECSYS","EQUINOX ", &
         "EXPOSURE","DATE-OBS","DATE-END","OBSERVER","ATT_SRC "/) )
    else
       !for tables, rename WCS keywords:
       call cp_attr(set1,set2,arr, &
         (/ "TELESCOP","DATAMODE","FILTER  ","OBS_MODE","OBS_ID  ","EXP_ID  ", &
            "EXPIDSTR","ORIGIN  ","OBJECT  ","RA_PNT  ","DEC_PNT ","PA_PNT  ", &
            "RADECSYS","EQUINOX ","EXPOSURE","DATE-OBS","DATE-END","OBSERVER", &
            "ATT_SRC ","SUBMODE " /) )
       call cp_wcs2table(set1,set2,"real32",arr, &
          (/ "CRPIX1  ","CRPIX2  ","CRVAL1  ","CRVAL2  ","CDELT1  ", &
             "CDELT2  "/) )
       call cp_wcs2table(set1,set2,"string",arr, (/ "CTYPE1  ","CTYPE2  "/) )
     end if

    if (mod > 2) then
      num_instr0=0
      num_image0=0
      do j=1,num_instr
        j0=expip(j)
        num_instr0=max(j0,num_instr0) 
        num_image0(j0)=num_image(j)
      end do
      n_instr=0 
      do j0=1,num_instr0
        if (num_image0(j0)>0) n_instr=n_instr+1
      end do  
  
      call setAttribute(set2,"N_INSTR",n_instr," ","Number of epic cameras")
      call setAttribute (set2,"PN_BNDS",num_image0(1), " ", &
                                               "Number of PN energy bands")
      call setAttribute (set2,"M1_BNDS",num_image0(2), " ", &
                                               "Number of MOS1 energy bands")
      call setAttribute (set2,"M2_BNDS",num_image0(3), " ", &
                                               "Number of MOS2 energy bands")

      if (mod == 4) then
        xidflg=.false.
        do j=1,num_instr
          if (xid_arr(j) > 0) xidflg=.true.
        end do
        if (xidflg) then
          call setAttribute (set2,"XID_BND",1," ", &
                                         "1 if XID band output is present")
        else
          call setAttribute (set2,"XID_BND",0," ", &
                                         "1 if XID band output is present")
        end if
      end if


      call setAttribute(set2,"FILT_PN",pnfilt,"EPN FILTER ID")
      call setAttribute(set2,"FILT_M1",m1filt,"EMOS1 FILTER ID")
      call setAttribute(set2,"FILT_M2",m2filt,"EMOS2 FILTER ID")
      call setAttribute(set2,"SUBM_PN",pnsub,"EPN SUBMODE")
      call setAttribute(set2,"SUBM_M1",m1sub,"EMOS1 SUBMODE")
      call setAttribute(set2,"SUBM_M2",m2sub,"EMOS2 SUBMODE")
      call setAttribute(set2,"EXPOS_PN",pnexpos,"EPN EXPOSURE")
      call setAttribute(set2,"EXPOS_M1",m1expos,"EMOS1 EXPOSURE")
      call setAttribute(set2,"EXPOS_M2",m2expos,"EMOS2 EXPOSURE")


      do i=1,num_instr
        do j=1,num_image(i)
          jj=j
          if (mod == 4) then
             if (xid_arr(i) == j) jj=9
          end if
          ipoint=ip(i,j)
          if (ipoint > 0) then
            write(strBand,'(I1)') jj 
            strEmin = strInst(expip(i))//"_"//strBand//"_ELO"
            strEmax = strInst(expip(i))//"_"//strBand//"_EHI"
            strComment = "E_min for Instr "//strInst(expip(i))//" , band "//strBand   
            call setAttribute(set2,strEmin,pimin(ipoint),"PI chan",strComment)
            strComment = "E_max for Instr "//strInst(expip(i))//" , band "//strBand   
            call setAttribute(set2,strEmax,pimax(ipoint),"PI chan",strComment)
          endif
        end do
      end do

      

    end if

    call setAttribute(set2,"CONTENT", content, "Content of file")

    call release(arr)

  end subroutine cp_head

  !---------------------------------------------------------------------
  subroutine cp_wcs2table (set1, set2, dtype, arr, attr)
  !---------------------------------------------------------------------
    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT), intent(IN) :: arr
    character(len=*), dimension(:), intent(IN) :: attr
    character(len=*), intent(IN) :: dtype
    character(len=8) attname
    real value
    character(len=80) svalue,comment 
    integer i

    do i=1,size(attr)
      comment="copy of WCS keyword "//attr(i)//" from input image" 
      if (dtype == "real32") then 
        if (hasAttribute(set1,attr(i))) then
          value=real32Attribute(set1,attr(i))
          attname = "T"//attr(i)
          call setAttribute(set2,attname,value,"",trim(comment))
        else 
          if (hasAttribute(arr,attr(i))) then
            value=real32Attribute(arr,attr(i))
            attname = "T"//attr(i)
            call setAttribute(set2,attname,value,"",trim(comment))
          end if
        end if
      elseif (dtype == "string") then
        if (hasAttribute(set1,attr(i))) then
          svalue=stringAttribute(set1,attr(i))
          attname = "T"//attr(i)
          call setAttribute(set2,attname,trim(svalue),trim(comment))
        else 
          if (hasAttribute(arr,attr(i))) then
            svalue=stringAttribute(arr,attr(i))
            attname = "T"//attr(i)
            call setAttribute(set2,attname,trim(svalue),trim(comment))
          end if
        end if 
      end if 
    end do

  end subroutine cp_wcs2table

  !--------------------------------------------------------------------
  subroutine get_files (param_name, imag, instrume, band, first)
  !---------------------------------------------------------------------
 
    character(len=*), intent(IN) :: param_name
    logical, intent(IN) :: first
    character(len=1024), dimension(*), intent(OUT) :: imag
    character(len=19), dimension(*), intent(OUT) :: instrume
    integer, dimension(*), intent(OUT) :: band
    character(len=80) errMess
    character(len=8) :: filter

    integer :: i, j, ib, ban
    integer :: nimage, ipoint
!    character(len=7) :: teles, instru, expid  
!    character(len=16),dimension(max_ima) :: ins
     character(len=3) :: teles
     character(len=5) :: instru
     character(len=11) :: expid
     character(len=19),dimension(max_ima) :: ins
    logical, dimension(max_ima) :: double

    type(DataSetT) :: set
    type(ArrayT) :: arr

    nimage = parameterCount(param_name) 
    if (nimage == 0) then
      call error("MissingParameter", "Missing input file name")
    end if

 
    if (first) then
      pnfilt="UNDEF"
      m1filt="UNDEF"    
      m2filt="UNDEF" 
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)

        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if


        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if


        if (hasAttribute(set,"FILTER")) then
          filter = stringAttribute(set,"FILTER")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"FILTER")) then
            filter = stringAttribute(arr,"FILTER")
          else
            call warning("MissingAttribute","Keyword FILTER is missing")
          end if
          call release(arr)
        end if


        if (index(instru,"PN") > 0 .or. &
            index(instru,"pn") > 0) then 
          pnfilt=filter
        else if (index(instru,"MOS1") > 0 .or. &
                 index(instru,"mos1") > 0) then
          m1filt=filter
        else if (index(instru,"MOS2") > 0 .or. &
                 index(instru,"mos2") > 0) then
          m2filt=filter
        end if
 
        


        if (hasAttribute(set,"EXP_ID")) then
          expid = stringAttribute(set,"EXP_ID")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"EXP_ID")) then
            expid = stringAttribute(arr,"EXP_ID")
          else
            expid = "---"
            call warning("MissingAttribute","Keyword EXP_ID is missing")
          end if
          call release(arr)
        end if
          
        instrume(i) = trim(teles) // trim(instru) // trim(expid)
 
        if (hasAttribute(set,"BAND")) then
          band(i) = int32Attribute(set, "BAND")
        else
          band(i) = 7
        end if
        call release(set)
      end do

      num_instr = 0
      double = .false.    
      ip = -1
      do i=1,nimage        
        if (.not. double(i)) num_instr = num_instr+1
        ib = 1 
        do while (ip(num_instr,ib) /= -1)
          ib = ib+1
        end do
        ip(num_instr,ib) = i
        num_image(num_instr) = ib
        do j=i+1,nimage
          if (instrume(i) == instrume(j)) double(j) = .true.
        end do
      end do

      do i=1,num_instr
        ipoint=ip(i,1) 
        if (index(instrume(ipoint),"PN") > 0 .or. &
            index(instrume(ipoint),"pn") > 0) then 
          expip(i) = 1
        else if (index(instrume(ipoint),"MOS1") > 0 .or. &
                 index(instrume(ipoint),"mos1") > 0) then
          expip(i) = 2
        else if (index(instrume(ipoint),"MOS2") > 0 .or. &
                 index(instrume(ipoint),"mos2") > 0) then
          expip(i) = 3
        else
          expip(i) = 4
          call error("WrongInst","Unknown instrument")
        end if
      end do
        
    else
      
      if (param_name(1:3) == "det") then
        if (nimage /= num_instr) then
           call error("FileMismatch", &
             "# detector masks /= # instruments")
        end if
      else 
        if (nimage /= nim) then
          call error("FileMismatch","Inconsistent number of input images")
        end if
      end if
 
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)

        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"EXP_ID")) then
          expid = stringAttribute(set,"EXP_ID")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"EXP_ID")) then
            expid = stringAttribute(arr,"EXP_ID")
          else
            expid = "---"
            call warning("MissingAttribute","Keyword EXP_ID is missing")
          end if
          call release(arr)
        end if

        ins(i) = trim(teles) // trim(instru) // trim(expid)


        call release(set)

      end do

      if (param_name(1:3) == "det") then 
         do i=1,num_instr
           do j=1,num_image(i)
             ipoint=ip(i,j)
             if (ins(i) /= instrume(ipoint)) then
               errMess = ": Wrong detector mask" 
               call error ("FileMismatch",errMess) 
             end if
           end do
         end do
      end if 

    end if

    nim = ip(num_instr,num_image(num_instr))

    

  end subroutine get_files 

  !---------------------------------------------------------------------
  subroutine co_image (fname, start, step, npix, exposure, xcen, ycen, &
                       ra_point, dec_point, pa_point, datetime, &
                       crpix1, crpix2, crval1, crval2, &
                       crdelt1,crdelt2,crota2, &
                       xoff,yoff,inst,ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix
    real(kind=double), dimension(2), intent(out) :: start, step 
    real(kind=double), intent(inout) :: ra_point, dec_point, pa_point, &
                       crpix1, crpix2, crval1, crval2, &
                       crdelt1, crdelt2, crota2 
    character(len=20), intent(out) :: datetime
    character(len=8) :: keyname
    character(len=2) :: istring
    real(kind=double) ::cos_pa, sin_pa, rr, dr, r0r, d0r, &
                       dra, ddec, xr, yr, dist1, dist2, del, &
                       crpix1old,crpix2old,crval1old,crval2old, &
                       crdelt1old, crdelt2old
 
    real(kind=single) :: exposure, xcen, ycen, r4huge, ran
    real(kind=double) :: r8huge
    integer, dimension(3) :: ixoff, iyoff
    integer :: i, ix, iy, iix, iiy, atype, itot, nev, ichip
    integer, dimension(3) :: nchips=(/12,7,7/)
    integer, dimension(2), save :: npix_save
    logical, save :: first = .true.

    integer, save :: iseed

    integer(kind=int32), dimension(:,:), pointer :: i4arr
    integer(kind=int16), dimension(:,:), pointer :: i2arr
    integer(kind=int8), dimension(:,:), pointer :: i1arr
    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=double), dimension(:,:), pointer :: r8arr
    integer, dimension(:), pointer :: shape

    type(datasett) :: set
    type(arrayt) :: arr
 
    r4huge=huge(1)
    r8huge=huge(1)

!    del = 1d-5
     del = 1d-3


    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)

    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    if (first) then
      pnsub='UNDEFINED'
      m1sub='UNDEFINED'
      m2sub='UNDEFINED'
      pnexpos=0.0
      m1expos=0.0
      m2expos=0.0
      npix_save=npix
      allocate (image(nim,npix(1),npix(2)))
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    itot=0
    atype = arrayDataType(arr)
     if (atype == Integer8) then
       i1arr => int8array2data(arr)
       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
             image(ipoint,ix,iy) = i1arr(iix,iiy)
             itot=itot+i1arr(iix,iiy)
         end do
       end do 
     else if (atype == Integer16) then
       i2arr => int16array2data(arr)
       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
           image(ipoint,ix,iy) = i2arr(iix,iiy)
           itot=itot+i2arr(iix,iiy)
         end do
       end do 
     else if (atype == Integer32) then 
       i4arr => int32array2data(arr)
       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
           image(ipoint,ix,iy) = i4arr(iix,iiy)
           itot=itot+i4arr(iix,iiy)
         end do
       end do 
     else if (atype == Real32) then
       r4arr => Real32array2data(arr)
       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
           image(ipoint,ix,iy) = nint(min(r4arr(iix,iiy),r4huge))
           itot=itot+r4arr(iix,iiy)
         end do
       end do 
     else if (atype == Real64) then
       r8arr => Real64array2data(arr)
       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
           image(ipoint,ix,iy) = nint(min(r8arr(iix,iiy),r8huge))
           itot=itot+r8arr(iix,iiy)
         end do
       end do 
     else
       call error("WrongType","Input image has wrong type")
     end if

! this is for testing ---------------------------------
!     do ix=1,npix(1)
!       do iy=1,npix(2)
!         image(ipoint,ix,iy)=0
!       end do
!     end do 
!     nev=npix(1)*npix(2)*0.2
!     !if (ipoint==1) iseed=-123456789 
!     do i=1,nev
!       ix=ran(iseed)*npix(1)+1
!       iy=ran(iseed)*npix(2)+1
!       if (ix.ge.1.and.ix.le.npix(1).and. &
!           iy.ge.1.and.iy.le.npix(2)) &
!         image(ipoint,ix,iy)=image(ipoint,ix,iy)+1
!     end do
!------------------------------------------------------

!    if (itot==0) call error("EmptyImage","Input image is empty")

    start(1) = 1d0
    start(2) = 1d0
    step(1) = 1d0
    step(2) = 1d0

    if (hasAttribute(set,"EXPOSURE")) then
      exposure = real32Attribute(set,"EXPOSURE")
      if (inst == 1) pnexpos=exposure
      if (inst == 2) m1expos=exposure
      if (inst == 3) m2expos=exposure  
    else if (hasAttribute(arr,"EXPOSURE")) then
      exposure = real32Attribute(arr,"EXPOSURE")
      if (inst == 1) pnexpos=exposure
      if (inst == 2) m1expos=exposure
      if (inst == 3) m2expos=exposure           
    else
      exposure = 1.0
      call warning("MissingAttribute", &
                   "No exposure keyword present: set to 1.0")
    end if 

    if (hasAttribute(set,"SUBMODE")) then
      submode = stringAttribute(set,"SUBMODE")
    else if (hasAttribute(arr,"SUBMODE")) then
      submode = stringAttribute(arr,"SUBMODE")
    else 
      submode="PrimeFullWindow"
    end if 
    if (inst == 1) pnsub=submode
    if (inst == 2) m1sub=submode
    if (inst == 3) m2sub=submode     




 
    if (hasAttribute(set,"DATE-OBS")) then
      datetime = stringAttribute(set,"DATE-OBS")
    else if (hasAttribute(arr,"DATE-OBS")) then
      datetime = stringAttribute(arr,"DATE-OBS")
    else 
      datetime=" "
    end if 

    if (hasAttribute(set,"RA_PNT")) then
      ra_point = real64Attribute(set,"RA_PNT")
    else  
      if (hasAttribute(arr,"RA_PNT")) then
        ra_point = real64Attribute(arr,"RA_PNT")
      else
        ra_point = 0d0
        call warning("MissingAttribute","Keyword RA_PNT is missing")
      end if
    end if

    if (hasAttribute(set,"DEC_PNT")) then
      dec_point = real64Attribute(set,"DEC_PNT")
    else  
      if (hasAttribute(arr,"DEC_PNT")) then
        dec_point = real64Attribute(arr,"DEC_PNT")
      else
        dec_point = 0d0
        call warning("MissingAttribute","Keyword DEC_PNT is missing")
      end if
    end if

    if (hasAttribute(set,"PA_PNT")) then
      pa_point = real64Attribute(set,"PA_PNT")
    else  
      if (hasAttribute(arr,"PA_PNT")) then
        pa_point = real64Attribute(arr,"PA_PNT")
      else
        pa_point = 0d0
        call warning("MissingAttribute","Keyword PA_PNT is missing")
      end if
    end if

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if
 
    if (hasAttribute(arr,"CROTA2")) then
       crota2 = real64Attribute(arr,"CROTA2")
    else
      if (hasAttribute(set,"CROTA2")) then
        crota2 = real64Attribute(set,"CROTA2")
      else
        crota2 = 0d0
        !call warning("MissingAttribute","Keyword CROTA2 is missing")
      end if
    end if

    if (crdelt1 > 0.0 .and. crdelt2 > 0.0) then
      dist1=0.36d0/abs(crdelt1)
      dist2=0.36d0/abs(crdelt2)
    else
      dist1=320.0d0
      dist2=320.0d0
    end if
    if (crpix1 < -dist1 .or. crpix1 > npix(1)+dist2 .or. &
        crpix2 < -dist2 .or. crpix2 > npix(2)+dist2) then
      call error("WrongRefPix","Reference pixel is outside of FOV")
    end if

    cos_pa = dcos(crota2 * pi/180d0)
    sin_pa = dsin(crota2 * pi/180d0)
    rr = ra_point * pi/180d0
    dr = dec_point * pi/180d0  
    r0r = crval1 * pi/180d0
    d0r = crval2 * pi/180d0
    call spher2(rr, dr, r0r, d0r, dra, ddec)
    xr =   dra * cos_pa + ddec * sin_pa ! Vorzeichen ??
    yr = - dra * sin_pa + ddec * cos_pa ! 
    xcen = xr * 180d0/pi / crdelt1 + crpix1
    ycen = yr * 180d0/pi / crdelt2 + crpix2

    if (xcen < 1.0 .or. xcen > npix(1) .or. &
        ycen < 1.0 .or. ycen > npix(2)) then
      xcen=min(max(xcen,1.0),real(npix(1)))
      ycen=min(max(ycen,1.0),real(npix(2)))
      call warning("WrongPointDir","Pointing direction is outside of image")
    end if

! read ontime values from image header:
    
    do ichip=1,nchips(inst)
      write(istring,'(I2.2)') ichip
      keyname="ONTIME"//istring
      if (hasAttribute(arr,keyname)) then
        ontime(ichip,ipoint) = real64Attribute(arr,keyname)  
      else
        if (hasAttribute(set,keyname)) then
          ontime(ichip,ipoint) = real64Attribute(set,keyname) 
        endif          
      endif  
  
    enddo
    


    call release(arr)
    call release(set)

  end subroutine co_image

  !---------------------------------------------------------------------
  subroutine co_bgima (fname, npix, &
                       crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                       xoff, yoff, inst, ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix 
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 
    integer, dimension(2), save :: npix_save
    integer, dimension(3) :: ixoff, iyoff
    integer :: ix, iy, iix, iiy, i, l
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=single) :: tot, bmax
    integer, dimension(:), pointer :: shape
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del

    type(datasett) :: set
    type(arrayt) :: arr

    del = 1d-5

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)
    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    if (first) then
      npix_save=npix
      allocate (bg_image(npix(1),npix(2),nim))
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    tot=0.
    bmax=0.
    do iy = 1,npix(2)
      do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1) &
        !bg_image(ix,iy,ipoint) = 0.2
        bg_image(ix,iy,ipoint) = r4arr(iix,iiy)
        tot=tot+r4arr(iix,iiy)
        bmax=max(r4arr(iix,iiy),bmax)
      end do
    end do 
     
    if (tot<=1e-3) call error("EmptyBgMap","Background map is empty") 

     !do iy = 1,npix(2)
     !  do ix = 1,npix(1)
     !    if (bg_image(ix,iy,ipoint)<2e-5) bg_image(ix,iy,ipoint)=bmax
     !  end do
     !end do

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if

    call release(arr)
    call release(set)
  
  end subroutine co_bgima

  !---------------------------------------------------------------------
  subroutine co_expima (fname, npix, &
                        crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                        xoff, yoff, inst, ipoint, &
                        xcen, ycen, exp_center, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), intent(inout) :: xcen, ycen, exp_center
    real(kind=single), dimension(3), intent(in) :: xoff, yoff   
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 
    integer, dimension(2), intent(out) :: npix 

    integer, dimension(2), save :: npix_save
    integer, dimension(3) :: ixoff, iyoff
    integer :: i, ix, iy, iix, iiy
    real(kind=single) :: tot, expmax
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    integer, dimension(:), pointer :: shape
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del

    type(datasett) :: set
    type(arrayt) :: arr

    del = 1d-5

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)

    if (hasAttribute(set,"XCEN")) then
      xcen = real32Attribute(set,"XCEN")
    else if (hasAttribute(arr,"XCEN")) then
      xcen = real32Attribute(arr,"XCEN")
    end if
    if (hasAttribute(set,"YCEN")) then
      ycen = real32Attribute(set,"YCEN")
    else if (hasAttribute(arr,"YCEN")) then
      ycen = real32Attribute(arr,"YCEN")
    end if
    exp_center=0.
    if (hasAttribute(set,"EXPOSURE")) then
      exp_center = real32Attribute(set,"EXPOSURE")
    else if (hasAttribute(arr,"EXPOSURE")) then
      exp_center = real32Attribute(arr,"EXPOSURE")
    end if 

    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    if (first) then
!     print *,'allocate exp image'
      npix_save=npix
      allocate (exp_image(npix(1),npix(2),nim))
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    tot=0.
    expmax=0.
    do iy = 1,npix(2)
      do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
        exp_image(ix,iy,ipoint) = r4arr(iix,iiy)
        tot=tot+r4arr(iix,iiy)
        expmax=max(expmax,r4arr(iix,iiy))
      end do
    end do 
      
    if (tot<=1e-3) call error("EmptyExpMap","Exposure map is empty")
    if (exp_center<=0.) exp_center=expmax

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if
    OotCorrDone(ipoint)=.false.
    if (hasAttribute(arr,"OOTCORR")) then
       OotCorrDone(ipoint)=booleanAttribute(arr,"OOTCORR")
    else
      if (hasAttribute(set,"OOTCORR")) then
        OotCorrDone(ipoint)=booleanAttribute(set,"OOTCORR")
      end if
    end if     


    call release(arr)
    call release(set)

  end subroutine co_expima

  !---------------------------------------------------------------------
  subroutine co_mask (fname, npix, &
                      crpix1, crpix2, crval1, crval2, crdelt1,crdelt2, &
                      xoff, yoff, inst, ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    real(kind=double), intent(inout) :: crpix1, crpix2, crval1, crval2, &
                                     crdelt1, crdelt2 
    integer, dimension(2), intent(out) :: npix

    integer, dimension(2), save :: npix_save
    integer, dimension(3) :: ixoff, iyoff
    integer :: i, ix, iy, iix, iiy, atype, itot
    logical, save :: first = .true.

    integer, dimension(:,:), pointer :: i4arr
    integer, dimension(:), pointer :: shape
    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=double) :: crpix1old,crpix2old,crval1old,crval2old, &
                         crdelt1old, crdelt2old, del

    type(datasett) :: set
    type(arrayt) :: arr
 
    del = 1d-5

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    if (first) then
!     print *,'allocate mask image:', nim,npix
      npix_save=npix
      allocate (mask_image(nim,npix(1),npix(2)))
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    itot=0
    atype = arrayDataType(arr)
    if (atype == Integer32) then
     i4arr => int32array2data(arr)
     do iy = 1,npix(2)
       do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1)  then
          if (i4arr(iix,iiy) == 0) then
            mask_image(ipoint,ix,iy) = .false.
          else
            mask_image(ipoint,ix,iy) = .true.
            itot=itot+1
          end if
        end if
       end do
     end do 
    end if
    if (atype == Real32) then
     r4arr => Real32array2data(arr)
     do iy = 1,npix(2)
       do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1)  then
          if (r4arr(iix,iiy) < 1e-6) then
            mask_image(ipoint,ix,iy) = .false.
          else
            mask_image(ipoint,ix,iy) = .true.
            itot=itot+1
          end if
        end if
       end do
     end do 
    end if

    if (itot==0) call error("EmptyMask","Detection mask is empty") 

    crpix1old=crpix1
    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
      if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
        if (crpix1old>-1d30.and.(crpix1<crpix1old-del.or.crpix1>crpix1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    crpix2old=crpix2
    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
      if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
        if (crpix2old>-1d30.and.(crpix2<crpix2old-del.or.crpix2>crpix2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    crval1old=crval1
    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
      if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
        if (crval1old>-1d30.and.(crval1<crval1old-del.or.crval1>crval1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        crval1 = 0d0
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    crval2old=crval2
    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
      if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
        if (crval2old>-1d30.and.(crval2<crval2old-del.or.crval2>crval2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    crdelt1old=crdelt1
    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
      if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
        if (crdelt1old>-1d30.and.(crdelt1<crdelt1old-del.or.crdelt1>crdelt1old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    crdelt2old=crdelt2
    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
      if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
         call error("FileMissmatch","Inconsistent WCS keywords")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
        if (crdelt2old>-1d30.and.(crdelt2<crdelt2old-del.or.crdelt2>crdelt2old+del)) &
           call error("FileMissmatch","Inconsistent WCS keywords")
      else
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if

    call release(arr)
    call release(set)

  end subroutine co_mask

!---------------------------------------------------------------------------------------

    subroutine impix2equ(xima,yima,ra,dec)

    real, intent(in) :: xima,yima
    real(kind=double), intent(out) :: ra,dec
    real(kind=double) :: xr,yr,dra,ddec,r0r,d0r,rr,dr,sin_pa,cos_pa
    real(kind=double) :: pi=3.141592654

        cos_pa = dcos(crota2 * pi/180d0)
        sin_pa = dsin(crota2 * pi/180d0)
        xr = (xima-crpix1) * crdelt1 * pi/180d0
        yr = (yima-crpix2) * crdelt2 * pi/180d0
        dra =  xr * cos_pa - yr * sin_pa
        ddec = xr * sin_pa + yr * cos_pa
        r0r = crval1 * pi/180d0
        d0r = crval2 * pi/180d0
        call antisp(dra,ddec,r0r,d0r,rr,dr)
        ra = rr * 180d0/pi
        dec = dr *180d0/pi

    
    end subroutine impix2equ


  !----------------------------------------------------------------------
  subroutine rawcoords(ra,dec,nsrc,datetime, &
                       chipix,chipiy,ccdinr,offax,azim)
  !----------------------------------------------------------------------

    use cal

    real(kind=double), dimension(max_fit), intent(in) :: ra, dec
    character(len=20), intent(in) :: datetime
    integer, intent(in) :: nsrc 
    integer, dimension(max_ins,max_fit), intent(out) :: & 
             chipix, chipiy, ccdinr
    
    real(kind=double), dimension(:), pointer :: theta, phi, &
!!!    real(kind=single), dimension(:), pointer :: theta, phi, &
                       cc2x, cc2y, cc2z, cc1x, cc1y, cc1z      
    real(kind=double) :: timestamp
    integer(int16), dimension(:), pointer :: chipX, chipY
    integer(int8), dimension(:), pointer :: ccdnr  
    integer :: i, j, j0

    type(SpacecraftAttitudeType) :: myattitude  
    type (BoreSightT) :: boresight

    type(EquatorialDirectionType), allocatable :: srcDirs(:)
    real(kind=double), dimension(max_ins,max_fit), intent(out) :: offax,azim 


    allocate(srcdirs(nsrc))


    do i=1,nsrc
      srcDirs(i)%ra = ra(i)*pi/180d0
      srcDirs(i)%dec = dec(i)*pi/180d0
    end do

    timeStamp = toTimeTag(datetime) ! in caloalutils/caloaldefs.f90        
!    boresight = CAL_getBoresight(TimeStamp)

    


    do j=1,num_instr
      myattitude%ra=ra_pnt(j)*pi/180d0
      myattitude%dec=dec_pnt(j)*pi/180d0
      myattitude%apos=pa_pnt(j)*pi/180d0       
      j0=expip(j)
      if (j0==1) call CAL_setState(EPN)
      if (j0==2) call CAL_setState(EMOS1)
      if (j0==3) call CAL_setState(EMOS2) 
      boresight = CAL_getBoresight(TimeStamp)
      call CAL_withSacCoordOrientation(boresight, myAttitude)
      call CAL_projectOntoTelCoord(boresight, srcDirs, theta, phi)
      call CAL_telCoordToCamCoord2(theta, phi, cc2x, cc2y, cc2z)
      call CAL_camCoord2ToCamCoord1(cc2x, cc2y, cc2z, cc1x, cc1y, cc1z)
      call CAL_camCoord1ToChipCoord(cc1x, cc1y, cc1z, chipX, chipY, ccdNr)
      do i=1,nsrc
        chipix(j,i)=chipX(i-1)
        chipiy(j,i)=chipY(i-1)
        ccdinr(j,i)=ccdNr(i-1)
        offax(j,i)=theta(i-1)
        azim(j,i)=phi(i-1)
      end do
    end do
    deallocate(srcdirs)
   
  end subroutine rawcoords




  end module edetect_io_ml






